/////////////
// GLOBALS //
/////////////
cbuffer MatrixBuffer
{
    matrix worldMatrix;
    matrix viewMatrix;
    matrix projectionMatrix;
};

cbuffer CameraBuffer
{
    float3 cameraPosition;
    float padding;
};

cbuffer SunLightBuffer
{
    float4 ambientColor;
    float4 diffuseColor;
    float3 lightDirection;
    float intensity;
};

cbuffer LightViewMatrixBuffer : register(b2)
{
    matrix lightViewMatrix;
    matrix lightProjectionMatrix;
};

//////////////
// TYPEDEFS //
//////////////
struct VertexInputType
{
    float4 position : POSITION;
    float2 tex : TEXCOORD0;
    float3 normal : NORMAL;
};

struct PixelInputType
{
    float4 position : SV_POSITION;
    float2 tex : TEXCOORD0;
    float3 normal : NORMAL;
    float4 lightPosition : TEXCOORD1;
};

////////////////////////////////////////////////////////////////////////////////
// Vertex Shader
////////////////////////////////////////////////////////////////////////////////
PixelInputType SunLightVertexShader(VertexInputType input)
{
    PixelInputType output;
    float4 worldPosition; // Dclaration de la variable manquante

    // Change the position vector to be 4 units for proper matrix calculations.
    input.position.w = 1.0f;
    
    // Calculate the position of the vertex against the world matrix
    worldPosition = mul(input.position, worldMatrix);
    
    // Calculate position in light view space.
    output.lightPosition = mul(worldPosition, lightViewMatrix);
    output.lightPosition = mul(output.lightPosition, lightProjectionMatrix);

    // Standard position calculation
    output.position = mul(worldPosition, viewMatrix);
    output.position = mul(output.position, projectionMatrix);
    
    // Store the texture coordinates for the pixel shader.
    output.tex = input.tex;
    
    // Calculate the normal vector against the world matrix only.
    output.normal = mul(input.normal, (float3x3)worldMatrix);
	
    // Normalize the normal vector.
    output.normal = normalize(output.normal);

    return output;
}