#pragma once

#include "Logger.h"
#include <d3d11.h>
#include <d3dcompiler.h>
#include <directxmath.h>
#include <fstream>
using namespace DirectX;
using namespace std;

class sunlight_shader_class
{

private :
	struct matrix_buffer_type
	{
		XMMATRIX world;
		XMMATRIX view;
		XMMATRIX projection;
	};

	struct camera_buffer_type
	{
		XMFLOAT3 camera_position;
		float padding;
	};

	struct sun_light_buffer_type
	{
		XMFLOAT4 diffuse_color;
		XMFLOAT4 ambient_color;
		XMFLOAT3 sun_direction;
		float intensity;
	};

	struct sun_light_color_buffer_type
	{
		XMFLOAT4 sun_color;
	};

	struct light_view_matrix_buffer_type
	{
		XMMATRIX light_view;
		XMMATRIX light_projection;
	};

public :
	sunlight_shader_class();
	sunlight_shader_class(const sunlight_shader_class&);
	~sunlight_shader_class();

	bool initialize(ID3D11Device*, HWND);
	void shutdown();
	bool render(
		ID3D11DeviceContext* device_context,
		int index_count,
		XMMATRIX world_matrix,
		XMMATRIX view_matrix,
		XMMATRIX projection_matrix,
		ID3D11ShaderResourceView* texture,
		XMFLOAT4 diffuse_color,
		XMFLOAT4 ambient_color,
		XMFLOAT3 sun_direction,
		float sun_intensity,
		ID3D11ShaderResourceView* shadow_map,
		XMMATRIX light_view_matrix,
		XMMATRIX light_projection_matrix
		);

private:
	bool initialize_shader(ID3D11Device*, HWND, WCHAR*, WCHAR*);
	void shutdown_shader();
	void output_shader_error_message(ID3D10Blob*, HWND, WCHAR*);

	bool set_shader_parameters(ID3D11DeviceContext*, XMMATRIX, XMMATRIX, XMMATRIX, ID3D11ShaderResourceView*, XMFLOAT4, XMFLOAT4, XMFLOAT3, float, ID3D11ShaderResourceView
	                           * shadowMap, XMMATRIX lightViewMatrix, XMMATRIX lightProjectionMatrix);
	void render_shader(ID3D11DeviceContext*, int);

private:
	ID3D11VertexShader* vertex_shader_;
	ID3D11PixelShader* pixel_shader_;
	ID3D11InputLayout* layout_;
	ID3D11SamplerState* sample_state_;
	ID3D11Buffer* matrix_buffer_;
	ID3D11Buffer* camera_buffer_;
	ID3D11Buffer* sunlight_buffer_;
	ID3D11Buffer* sunlight_color_buffer_;
	ID3D11Buffer* sunlight_position_buffer_;
	ID3D11Buffer* light_view_matrix_buffer_;
	ID3D11ShaderResourceView* shadow_map_texture_;
	ID3D11SamplerState* shadow_sampler_state_;
};