#include "sunlight_shader_class.h"

sunlight_shader_class::sunlight_shader_class()
{
	vertex_shader_ = nullptr;
	pixel_shader_ = nullptr;
	layout_ = nullptr;
	sample_state_ = nullptr;
	matrix_buffer_ = nullptr;
	camera_buffer_ = nullptr;
	sunlight_buffer_ = nullptr;
	sunlight_color_buffer_ = nullptr;
	sunlight_position_buffer_ = nullptr;
	light_view_matrix_buffer_ = nullptr;
	shadow_map_texture_ = nullptr;
	shadow_sampler_state_ = nullptr;
}

sunlight_shader_class::sunlight_shader_class(const sunlight_shader_class&)
{
}

sunlight_shader_class::~sunlight_shader_class()
{
}

bool sunlight_shader_class::initialize(ID3D11Device* device, HWND hwnd)
{
	bool result;
	WCHAR vs_filename[128];
	WCHAR ps_filename[128];

	// Dfinir les fichiers des shaders vertex et pixel
	wcscpy_s(vs_filename, 128, L"./src/hlsl/sunlight.vs");
	wcscpy_s(ps_filename, 128, L"./src/hlsl/sunlight.ps");

	// Initialiser les shaders
	result = initialize_shader(device, hwnd, vs_filename, ps_filename);
	if (!result)
	{
		Logger::Get().Log("chec d'initialisation des shaders sunlight", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	return true;
}

void sunlight_shader_class::shutdown()
{
	shutdown_shader();
	return;
}

bool sunlight_shader_class::render(ID3D11DeviceContext* device_context, int index_count, XMMATRIX world_matrix, XMMATRIX view_matrix, XMMATRIX projection_matrix, ID3D11ShaderResourceView* texture, XMFLOAT4 diffuse_color, XMFLOAT4 ambient_color, XMFLOAT3 sun_direction, float sun_intensity, ID3D11ShaderResourceView* shadow_map, XMMATRIX light_view_matrix, XMMATRIX light_projection_matrix)
{
	bool result;

	// Dfinir les paramtres du shader
	result = set_shader_parameters(device_context, world_matrix, view_matrix, projection_matrix, texture, diffuse_color, ambient_color, sun_direction, sun_intensity, shadow_map, light_view_matrix, light_projection_matrix);
	if (!result)
	{
		Logger::Get().Log("chec de dfinition des paramtres du shader sunlight", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Rendre le shader
	render_shader(device_context, index_count);

	return true;
}

bool sunlight_shader_class::initialize_shader(ID3D11Device* device, HWND hwnd, WCHAR* vs_filename, WCHAR* ps_filename)
{
	HRESULT result;
	ID3D10Blob* error_message = nullptr;
	ID3D10Blob* vertex_shader_buffer = nullptr;
	ID3D10Blob* pixel_shader_buffer = nullptr;
	D3D11_INPUT_ELEMENT_DESC polygon_layout[3];
	unsigned int num_elements;
	D3D11_BUFFER_DESC matrix_buffer_desc;
	D3D11_BUFFER_DESC camera_buffer_desc;
	D3D11_BUFFER_DESC light_buffer_desc;
	D3D11_BUFFER_DESC light_color_buffer_desc;
	D3D11_BUFFER_DESC light_view_matrix_buffer_desc;
	D3D11_SAMPLER_DESC sampler_desc;
	D3D11_SAMPLER_DESC shadow_sampler_desc;

	// Compiler le vertex shader
	result = D3DCompileFromFile(vs_filename, NULL, NULL, "SunLightVertexShader", "vs_5_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, &vertex_shader_buffer, &error_message);
	if (FAILED(result))
	{
		if (error_message)
		{
			output_shader_error_message(error_message, hwnd, vs_filename);
		}
		else
		{
			MessageBox(hwnd, vs_filename, L"Fichier shader manquant", MB_OK);
		}

		return false;
	}

	// Compiler le pixel shader
	result = D3DCompileFromFile(ps_filename, NULL, NULL, "SunLightPixelShader", "ps_5_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, &pixel_shader_buffer, &error_message);
	if (FAILED(result))
	{
		if (error_message)
		{
			output_shader_error_message(error_message, hwnd, ps_filename);
		}
		else
		{
			MessageBox(hwnd, ps_filename, L"Fichier shader manquant", MB_OK);
		}

		return false;
	}

	// Crer le vertex shader  partir du buffer
	result = device->CreateVertexShader(vertex_shader_buffer->GetBufferPointer(), vertex_shader_buffer->GetBufferSize(), NULL, &vertex_shader_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du vertex shader", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Crer le pixel shader  partir du buffer
	result = device->CreatePixelShader(pixel_shader_buffer->GetBufferPointer(), pixel_shader_buffer->GetBufferSize(), NULL, &pixel_shader_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du pixel shader", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Crer le layout d'entre du vertex shader
	polygon_layout[0].SemanticName = "POSITION";
	polygon_layout[0].SemanticIndex = 0;
	polygon_layout[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
	polygon_layout[0].InputSlot = 0;
	polygon_layout[0].AlignedByteOffset = 0;
	polygon_layout[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
	polygon_layout[0].InstanceDataStepRate = 0;

	polygon_layout[1].SemanticName = "TEXCOORD";
	polygon_layout[1].SemanticIndex = 0;
	polygon_layout[1].Format = DXGI_FORMAT_R32G32_FLOAT;
	polygon_layout[1].InputSlot = 0;
	polygon_layout[1].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
	polygon_layout[1].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
	polygon_layout[1].InstanceDataStepRate = 0;

	polygon_layout[2].SemanticName = "NORMAL";
	polygon_layout[2].SemanticIndex = 0;
	polygon_layout[2].Format = DXGI_FORMAT_R32G32B32_FLOAT;
	polygon_layout[2].InputSlot = 0;
	polygon_layout[2].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
	polygon_layout[2].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
	polygon_layout[2].InstanceDataStepRate = 0;

	// Obtenir le nombre d'lments dans le layout
	num_elements = sizeof(polygon_layout) / sizeof(polygon_layout[0]);

	// Crer le layout d'entre du vertex
	result = device->CreateInputLayout(polygon_layout, num_elements, vertex_shader_buffer->GetBufferPointer(), vertex_shader_buffer->GetBufferSize(), &layout_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du layout d'entre", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Librer les buffers des shaders
	vertex_shader_buffer->Release();
	vertex_shader_buffer = nullptr;

	pixel_shader_buffer->Release();
	pixel_shader_buffer = nullptr;

	// Configurer les descripteurs des buffers constants
	// Descripteur du buffer de matrices
	matrix_buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
	matrix_buffer_desc.ByteWidth = sizeof(matrix_buffer_type);
	matrix_buffer_desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	matrix_buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	matrix_buffer_desc.MiscFlags = 0;
	matrix_buffer_desc.StructureByteStride = 0;

	// Crer le buffer constant de matrices
	result = device->CreateBuffer(&matrix_buffer_desc, NULL, &matrix_buffer_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du buffer constant de matrices", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Descripteur du buffer de camra
	camera_buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
	camera_buffer_desc.ByteWidth = sizeof(camera_buffer_type);
	camera_buffer_desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	camera_buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	camera_buffer_desc.MiscFlags = 0;
	camera_buffer_desc.StructureByteStride = 0;

	// Crer le buffer constant de camra
	result = device->CreateBuffer(&camera_buffer_desc, NULL, &camera_buffer_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du buffer constant de camra", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Descripteur du buffer de lumire du soleil
	light_buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
	light_buffer_desc.ByteWidth = sizeof(sun_light_buffer_type);
	light_buffer_desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	light_buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	light_buffer_desc.MiscFlags = 0;
	light_buffer_desc.StructureByteStride = 0;

	// Crer le buffer constant de lumire du soleil
	result = device->CreateBuffer(&light_buffer_desc, NULL, &sunlight_buffer_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du buffer constant de lumire du soleil", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Descripteur du buffer de couleur de lumire du soleil
	light_color_buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
	light_color_buffer_desc.ByteWidth = sizeof(sun_light_color_buffer_type);
	light_color_buffer_desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	light_color_buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	light_color_buffer_desc.MiscFlags = 0;
	light_color_buffer_desc.StructureByteStride = 0;

	// Crer le buffer constant de couleur de lumire du soleil
	result = device->CreateBuffer(&light_color_buffer_desc, NULL, &sunlight_color_buffer_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du buffer constant de couleur de lumire du soleil", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Descripteur du buffer de matrices de vue de lumire
	light_view_matrix_buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
	light_view_matrix_buffer_desc.ByteWidth = sizeof(light_view_matrix_buffer_type);
	light_view_matrix_buffer_desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	light_view_matrix_buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	light_view_matrix_buffer_desc.MiscFlags = 0;
	light_view_matrix_buffer_desc.StructureByteStride = 0;

	// Crer le buffer constant de matrices de vue de lumire
	result = device->CreateBuffer(&light_view_matrix_buffer_desc, NULL, &light_view_matrix_buffer_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du buffer constant de matrices de vue de lumire", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Crer un sampler state pour la texture
	sampler_desc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	sampler_desc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
	sampler_desc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
	sampler_desc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
	sampler_desc.MipLODBias = 0.0f;
	sampler_desc.MaxAnisotropy = 1;
	sampler_desc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	sampler_desc.BorderColor[0] = 0;
	sampler_desc.BorderColor[1] = 0;
	sampler_desc.BorderColor[2] = 0;
	sampler_desc.BorderColor[3] = 0;
	sampler_desc.MinLOD = 0;
	sampler_desc.MaxLOD = D3D11_FLOAT32_MAX;

	// Crer le sampler state
	result = device->CreateSamplerState(&sampler_desc, &sample_state_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du sampler state", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Crer un sampler state pour la shadow map
	shadow_sampler_desc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
	shadow_sampler_desc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
	shadow_sampler_desc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
	shadow_sampler_desc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
	shadow_sampler_desc.MipLODBias = 0.0f;
	shadow_sampler_desc.MaxAnisotropy = 1;
	shadow_sampler_desc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	shadow_sampler_desc.BorderColor[0] = 1.0f;
	shadow_sampler_desc.BorderColor[1] = 1.0f;
	shadow_sampler_desc.BorderColor[2] = 1.0f;
	shadow_sampler_desc.BorderColor[3] = 1.0f;
	shadow_sampler_desc.MinLOD = 0;
	shadow_sampler_desc.MaxLOD = D3D11_FLOAT32_MAX;

	// Crer le sampler state pour la shadow map
	result = device->CreateSamplerState(&shadow_sampler_desc, &shadow_sampler_state_);
	if (FAILED(result))
	{
		Logger::Get().Log("chec de cration du sampler state pour la shadow map", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	return true;
}

void sunlight_shader_class::shutdown_shader()
{
	// Librer tous les objets
	if (shadow_sampler_state_)
	{
		shadow_sampler_state_->Release();
		shadow_sampler_state_ = nullptr;
	}

	if (light_view_matrix_buffer_)
	{
		light_view_matrix_buffer_->Release();
		light_view_matrix_buffer_ = nullptr;
	}

	if (sunlight_color_buffer_)
	{
		sunlight_color_buffer_->Release();
		sunlight_color_buffer_ = nullptr;
	}

	if (sunlight_buffer_)
	{
		sunlight_buffer_->Release();
		sunlight_buffer_ = nullptr;
	}

	if (camera_buffer_)
	{
		camera_buffer_->Release();
		camera_buffer_ = nullptr;
	}

	if (matrix_buffer_)
	{
		matrix_buffer_->Release();
		matrix_buffer_ = nullptr;
	}

	if (sample_state_)
	{
		sample_state_->Release();
		sample_state_ = nullptr;
	}

	if (layout_)
	{
		layout_->Release();
		layout_ = nullptr;
	}

	if (pixel_shader_)
	{
		pixel_shader_->Release();
		pixel_shader_ = nullptr;
	}

	if (vertex_shader_)
	{
		vertex_shader_->Release();
		vertex_shader_ = nullptr;
	}

	return;
}

void sunlight_shader_class::output_shader_error_message(ID3D10Blob* error_message, HWND hwnd, WCHAR* shader_filename)
{
	char* compile_errors;
	unsigned long long buffer_size;
	ofstream fout;

	// Obtenir un pointeur vers le texte d'erreur dans le buffer d'erreur
	compile_errors = (char*)(error_message->GetBufferPointer());

	// Obtenir la taille du message
	buffer_size = error_message->GetBufferSize();

	// Ouvrir un fichier pour crire les messages d'erreur
	fout.open("shader-error.txt");

	// crire le message d'erreur
	for (unsigned long long i = 0; i < buffer_size; i++)
	{
		fout << compile_errors[i];
	}

	// Fermer le fichier
	fout.close();

	// Librer le buffer d'erreur
	error_message->Release();
	error_message = nullptr;

	// Afficher un message d'erreur
	MessageBox(hwnd, L"Erreur de compilation du shader. Voir shader-error.txt pour plus de dtails.", shader_filename, MB_OK);

	return;
}

bool sunlight_shader_class::set_shader_parameters(ID3D11DeviceContext* device_context, XMMATRIX world_matrix, XMMATRIX view_matrix, XMMATRIX projection_matrix, ID3D11ShaderResourceView* texture, XMFLOAT4 diffuse_color, XMFLOAT4 ambient_color, XMFLOAT3 sun_direction, float sun_intensity, ID3D11ShaderResourceView* shadow_map, XMMATRIX light_view_matrix, XMMATRIX light_projection_matrix)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mapped_resource;
	matrix_buffer_type* data_ptr;
	camera_buffer_type* camera_data_ptr;
	sun_light_buffer_type* light_data_ptr;
	sun_light_color_buffer_type* light_color_data_ptr;
	light_view_matrix_buffer_type* light_view_data_ptr;

	// Transposer les matrices pour qu'elles soient prtes pour le shader
	world_matrix = XMMatrixTranspose(world_matrix);
	view_matrix = XMMatrixTranspose(view_matrix);
	projection_matrix = XMMatrixTranspose(projection_matrix);
	light_view_matrix = XMMatrixTranspose(light_view_matrix);
	light_projection_matrix = XMMatrixTranspose(light_projection_matrix);

	// Verrouiller le buffer constant pour l'criture
	result = device_context->Map(matrix_buffer_, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapped_resource);
	if (FAILED(result))
	{
		Logger::Get().Log("chec du verrouillage du buffer constant de matrices", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Obtenir un pointeur vers les donnes
	data_ptr = (matrix_buffer_type*)mapped_resource.pData;

	// Copier les matrices dans le buffer constant
	data_ptr->world = world_matrix;
	data_ptr->view = view_matrix;
	data_ptr->projection = projection_matrix;

	// Dverrouiller le buffer constant
	device_context->Unmap(matrix_buffer_, 0);

	// Dfinir la position du buffer constant dans le vertex shader
	unsigned int buffer_number = 0;

	// Dfinir le buffer constant du vertex shader
	device_context->VSSetConstantBuffers(buffer_number, 1, &matrix_buffer_);

	// Verrouiller le buffer constant de camra
	result = device_context->Map(camera_buffer_, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapped_resource);
	if (FAILED(result))
	{
		Logger::Get().Log("chec du verrouillage du buffer constant de camra", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Obtenir un pointeur vers les donnes
	camera_data_ptr = (camera_buffer_type*)mapped_resource.pData;

	// TODO: Ajouter la position de la camra ici si ncessaire
	camera_data_ptr->camera_position = XMFLOAT3(0.0f, 0.0f, 0.0f);
	camera_data_ptr->padding = 0.0f;

	// Dverrouiller le buffer constant
	device_context->Unmap(camera_buffer_, 0);

	// Dfinir la position du buffer constant dans le vertex shader
	buffer_number = 1;

	// Dfinir le buffer constant du vertex shader
	device_context->VSSetConstantBuffers(buffer_number, 1, &camera_buffer_);

	// Verrouiller le buffer constant de lumire
	result = device_context->Map(sunlight_buffer_, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapped_resource);
	if (FAILED(result))
	{
		Logger::Get().Log("chec du verrouillage du buffer constant de lumire", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Obtenir un pointeur vers les donnes
	light_data_ptr = (sun_light_buffer_type*)mapped_resource.pData;

	// Copier les paramtres de lumire dans le buffer constant
	light_data_ptr->ambient_color = ambient_color;
	light_data_ptr->diffuse_color = diffuse_color;
	light_data_ptr->sun_direction = sun_direction;
	light_data_ptr->intensity = sun_intensity;

	// Dverrouiller le buffer constant
	device_context->Unmap(sunlight_buffer_, 0);

	// Dfinir la position du buffer constant dans le pixel shader
	buffer_number = 0;

	// Dfinir le buffer constant du pixel shader
	device_context->PSSetConstantBuffers(buffer_number, 1, &sunlight_buffer_);
	device_context->VSSetConstantBuffers(buffer_number, 1, &sunlight_buffer_);

	// Verrouiller le buffer constant de matrices de vue de lumire
	result = device_context->Map(light_view_matrix_buffer_, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapped_resource);
	if (FAILED(result))
	{
		Logger::Get().Log("chec du verrouillage du buffer constant de matrices de vue de lumire", __FILE__, __LINE__, Logger::LogLevel::Error);
		return false;
	}

	// Obtenir un pointeur vers les donnes
	light_view_data_ptr = (light_view_matrix_buffer_type*)mapped_resource.pData;

	// Copier les matrices de vue de lumire dans le buffer constant
	light_view_data_ptr->light_view = light_view_matrix;
	light_view_data_ptr->light_projection = light_projection_matrix;

	// Dverrouiller le buffer constant
	device_context->Unmap(light_view_matrix_buffer_, 0);

	// Dfinir la position du buffer constant dans le vertex shader
	buffer_number = 2;

	// Dfinir le buffer constant du vertex shader
	device_context->VSSetConstantBuffers(buffer_number, 1, &light_view_matrix_buffer_);

	// Dfinir la ressource de texture du shader
	device_context->PSSetShaderResources(0, 1, &texture);

	// Dfinir la ressource de shadow map du shader
	device_context->PSSetShaderResources(1, 1, &shadow_map);

	// Dfinir les sampler states dans le pixel shader
	device_context->PSSetSamplers(0, 1, &sample_state_);
	device_context->PSSetSamplers(1, 1, &shadow_sampler_state_);

	return true;
}

void sunlight_shader_class::render_shader(ID3D11DeviceContext* device_context, int index_count)
{
	// Dfinir le layout du vertex input
	device_context->IASetInputLayout(layout_);

	// Dfinir les shaders vertex et pixel
	device_context->VSSetShader(vertex_shader_, NULL, 0);
	device_context->PSSetShader(pixel_shader_, NULL, 0);

	// Rendre la gomtrie
	device_context->DrawIndexed(index_count, 0, 0);

	return;
}